((
$Date: 2008-09-27 18:19:05 +0200 (sb, 27 sep 2008) $
$Revision: 586 $
$Author: astrorafael $
$HeadURL: svn://localhost/repos/vfx/forthql/tags/forthql-1.0/src/sqlite3.fs $

2008 By Rafael Gonzalez Fuentetaja.
This file is placed in the public domain. NO WARRANTY.
))
\ ####################################
\ *! sqlite3
\ *R @node{SQLite3 Database Interface}
\ *T SQLite3 Database Interface
\ ####################################

\ ***************
\ *S Introduction
\ ***************

\ *P This module is a Forth DBMS driver to SQLite3.
\ ** It implements a minimal interface for use with *\i{ForthQL} and
\ ** an extended interface for SQL3Lite specific operations.

\ ==============
\ *N Portability
\ ==============

\ *P This module is written in *\i{VFX Forth for Linux} but it can probably
\ ** be ported easily to other Forth environments as well.

\ *P The following aspects of VFX has been used:
\ *(
\ *B C Callback Mechanism. 
\ *B Exception messages and codes.
\ *B Modules, encapsulating words into modules and exporting words.
\ *B MPEisms. Convenience words defined by MPE for its products
\ ** like *\fo{?comp}, *\fo{$CRLR}, *\fo{ALLOT&ERASE} or *\fo{?throw}
\ *)

\ ***********
\ *S Glossary
\ ***********

[undefined] sql-multithread [if]
0 Constant sql-multithread
\ *G Compilation option when including this module. 
\ ** All DBMS drivers should test & define this value.   
[endif]


MODULE SQLITE3
\ *G
   
[undefined] libsqlite3.so.0 [if]
   Library: libsqlite3.so.0
[endif]


\ ==============
\ *N C interface
\ ==============


\ -----------------------
\ *H SQLite3 return codes
\ -----------------------

\ *[
0 Constant SQLITE_OK	     \ Successful result
1 Constant SQLITE_ERROR      \ SQL error or missing database
2 Constant SQLITE_INTERNAL   \ An internal logic error in SQLite
3 Constant SQLITE_PERM	     \ Access permission denied
4 Constant SQLITE_ABORT	     \ Callback routine requested an abort
5 Constant SQLITE_BUSY	     \ The database file is locked
6 Constant SQLITE_LOCKED     \ A table in the database is locked
7 Constant SQLITE_NOMEM	     \ A malloc() failed
8 Constant SQLITE_READONLY   \ Attempt to write a readonly database
9 Constant SQLITE_INTERRUPT  \ Operation terminated by sqlite_interrupt()
#10 Constant SQLITE_IOERR    \ Some kind of disk I/O error occurred
#11 Constant SQLITE_CORRUPT  \ The database disk image is malformed
#12 Constant SQLITE_NOTFOUND \ (Internal Only) Table or record not found
#13 Constant SQLITE_FULL     \ Insertion failed because database is full
#14 Constant SQLITE_CANTOPEN \ Unable to open the database file
#15 Constant SQLITE_PROTOCOL \ Database lock protocol error
#16 Constant SQLITE_EMPTY    \ (Internal Only) Database table is empty
#17 Constant SQLITE_SCHEMA   \ The database schema changed
#18 Constant SQLITE_TOOBIG   \ Too much data for one row of a table
#19 Constant SQLITE_CONSTRAINT	    \ Abort due to contraint violation
#20 Constant SQLITE_MISMATCH \ Data type mismatch
#21 Constant SQLITE_MISUSE   \ Library used incorrectly
#22 Constant SQLITE_NOLFS    \ Uses OS features not supported on host
#23 Constant SQLITE_AUTH     \ Authorization denied

#100 Constant SQLITE_ROW     \ sqlite_step() has another row ready
#101 Constant SQLITE_DONE    \ sqlite_step() has finished executing
\ *]

EXPORT SQLITE_OK
EXPORT SQLITE_ERROR
EXPORT SQLITE_INTERNAL
EXPORT SQLITE_PERM
EXPORT SQLITE_ABORT
EXPORT SQLITE_BUSY
EXPORT SQLITE_LOCKED
EXPORT SQLITE_NOMEM
EXPORT SQLITE_READONLY
EXPORT SQLITE_INTERRUPT
EXPORT SQLITE_IOERR
EXPORT SQLITE_CORRUPT
EXPORT SQLITE_NOTFOUND
EXPORT SQLITE_FULL
EXPORT SQLITE_CANTOPEN
EXPORT SQLITE_PROTOCOL
EXPORT SQLITE_EMPTY
EXPORT SQLITE_SCHEMA
EXPORT SQLITE_TOOBIG
EXPORT SQLITE_CONSTRAINT
EXPORT SQLITE_MISMATCH
EXPORT SQLITE_MISUSE 
EXPORT SQLITE_AUTH

EXPORT SQLITE_ROW
EXPORT SQLITE_DONE


\ ---------------------
\ *H SQLite3 data types
\ ---------------------

\ *[
1 Constant SQLITE_INTEGER
2 Constant SQLITE_FLOAT
3 Constant SQLITE_TEXT
4 Constant SQLITE_BLOB
5 Constant SQLITE_NULL
\ *]

EXPORT SQLITE_INTEGER
EXPORT SQLITE_FLOAT
EXPORT SQLITE_TEXT
EXPORT SQLITE_BLOB
EXPORT SQLITE_NULL


\ --------------
\ +H C functions
\ --------------

\ +G The minimal C interface needed for ForthQL
\ +* and callback-style processing.

\ +[
Extern: int sqlite3_libversion_number(void);

Extern: const char* sqlite3_libversion(void);

Extern: int sqlite3_open( const char*, sqlite3** );

Extern: int sqlite3_close( sqlite3* );

Extern: const char* sqlite3_errmsg( sqlite3* );

Extern: int sqlite3_exec(
  sqlite3*,				
  const char*, 				
  cback*,					
  void*,				
  char**				
);

Extern: void sqlite3_free( void* );

\ SQLite3 specific For non-callback style processing

Extern: int sqlite3_prepare_v2(
  sqlite3*,
  const char*,
  int,      
  sqlite3_stmt**,
  const char**   
);

Extern: int sqlite3_step( sqlite3_stmt* );

Extern: int sqlite3_reset( sqlite3_stmt* );

Extern: int sqlite3_finalize( sqlite3_stmt* );

Extern: int sqlite3_column_count(sqlite3_stmt*);

Extern: int sqlite3_column_type(sqlite3_stmt*, int );

Extern: const char* sqlite3_column_decltype(sqlite3_stmt*, int );

Extern: const char* sqlite3_column_name(sqlite3_stmt*, int );

Extern: const unsigned char* sqlite3_column_text(sqlite3_stmt*, int );
\ +]


0 [if]					\ may be needed in the future ...
const void* sqlite3_column_blob( sqlite3_stmt*, int);
int sqlite3_column_bytes(sqlite3_stmt*, int);
int sqlite3_column_bytes16(sqlite3_stmt*, int );
int sqlite3_column_int(sqlite3_stmt*, int );
long long int sqlite3_column_int64(sqlite3_stmt*, int ); \ WILL THIS WORK ? 
double sqlite3_column_double(sqlite3_stmt*, int ); \ HOW TO INTERFACE DOUBLES ?
[then]



: .sql3-version
\ *G Print the SQLite3 version being loaded
\ ** The SQLite3 version number is an integer
\ ** with the value (X*1000000 + Y*1000 + Z).
   ." SQLite3 version being loaded: " sqlite3_libversion .z$ cr
;


' .sql3-version AtCold

cr .sql3-version

\ =================
\ *N Generic DB API
\ =================


\ -------------
\ *H Exceptions
\ -------------


ErrDef SQL3ColumnIndex "Index out of bounds in column widths array"
\ *G

ErrDef SQL3ColumnWidthError "Column specifier not within range"
\ *G

NextError @ Constant ErrorBase

   
\ *P The ones below comes from the SQlite3 DBMS itself. 
\ *[
ErrDef SQL3_ERROR      "SQLite3: SQL error or missing database"
ErrDef SQL3_INTERNAL   "SQLite3: An internal logic error in SQLite"
ErrDef SQL3_PERM       "SQLite3: Access permission denied"
ErrDef SQL3_ABORT      "SQLite3: Callback routine requested an abort"
ErrDef SQL3_BUSY       "SQLite3: The database file is locked"
ErrDef SQL3_LOCKED     "SQLite3: A table in the database is locked"
ErrDef SQL3_NOMEM      "SQLite3: A malloc() failed"
ErrDef SQL3_READONLY   "SQLite3: Attempt to write a readonly database"
ErrDef SQL3_INTERRUPT  "SQLite3: Operation terminated by sqlite_interrupt()"
ErrDef SQL3_IOERR      "SQLite3: Some kind of disk I/O error occurred"
ErrDef SQL3_CORRUPT    "SQLite3: The database disk image is malformed"
ErrDef SQL3_NOTFOUND   "SQLite3: (Internal Only) Table or record not found"
ErrDef SQL3_FULL       "SQLite3: Insertion failed because database is full"
ErrDef SQL3_CANTOPEN   "SQLite3: Unable to open the database file"
ErrDef SQL3_PROTOCOL   "SQLite3: Database lock protocol error"
ErrDef SQL3_EMPTY      "SQLite3: (Internal Only) Database table is empty"
ErrDef SQL3_SCHEMA     "SQLite3: The database schema changed"
ErrDef SQL3_TOOBIG     "SQLite3: Too much data for one row of a table"
ErrDef SQL3_CONSTRAINT "SQLite3: Abort due to contraint violation"
ErrDef SQL3_MISMATCH   "SQLite3: Data type mismatch"
ErrDef SQL3_MISUSE     "SQLite3: Library used incorrectly"
ErrDef SQL3_AUTH       "SQLite3: Authorization denied" 
\ *]

\ ------------------
\ +H State variables
\ ------------------

(( For DocGen use only
Variable sql3-handle
\ +G  Handle to a SQLite3 connection.
\ +* A *\fo{USER} variable when compiled with *\fo{sql-multithread} set to 1.

Variable sql3-sth
\ +G Prepared statement handle.
\ +* A *\fo{USER} variable when compiled with *\fo{sql-multithread} set to 1.
))
   
sql-multithread [if]   
   cell +User sql3-handle
   cell +User sql3-sth
[else]
   Variable sql3-handle
   Variable sql3-sth
[then]


sql3-handle off				\ only for the main task
sql3-sth off

\ ------------
\ *H Functions
\ ------------
   
: db-open			  \ ca1 u1 -- ior
\ *G Open a connection to a SQLite3 database file given by path *\i{ca1 u1}.
\ ** Returned handle is stored in *\fo{SQL3-HANDLE} *\fo{USER} variable.
\ ** Does nothing if already open (*\fo{SQL3-HANDLE} is non-zero).   
   { | tmp[ cell ] path[ #256 chars ]  -- }
   sql3-handle @ if
      2drop SQLITE_OK exit
   then
   PAD DUP >R PLACE R> path[ caddr>zaddr    \ temporary copy in PAD
   path[ tmp[ sqlite3_open
   tmp[ @ sql3-handle !
;
EXPORT db-open

: db-close			  \ -- ior
\ *G Close a connection to a SQLite3 database. Handle is taken from
\ ** *\fo{SQL3-HANDLE}. Clears *\fo{SQL3-HANDLE} as well.
\ ** Does nothing if already closed (*\fo{SQL3-HANDLE} is zero).    
   sql3-handle @ 0= if
      SQLITE_OK exit
   then
   sql3-handle @ sqlite3_close
   sql3-handle off		
;
EXPORT db-close

: db-errmsg			  \ ior -- ca1 u1
\ *G Get an error description message *\i{ca1 u1} from an *\i{ior}.
\ ** *\b{Warning:} In SQLite3, *\i{ior} is not looked up,
\ ** the most recent error produced is retrieved.    
   drop sql3-handle @ sqlite3_errmsg zcount
;
EXPORT db-errmsg


: db-execute			  \ z-addr -- ior
\ *G Execute SQL statement *\i{z-addr} not returning any output.
   { | errmsg[ cell ] -- ior }
   errmsg[ off
   sql3-handle @ swap NULL dup errmsg[ sqlite3_exec
   errmsg[ @ ?dup if sqlite3_free then
;
EXPORT db-execute


: db-process			  \ z-addr xt -- ior
\ *G Execute SQL statement *\i{z-addr}.
\ ** When complete, word given by *\i{xt} is repetidely called,
\ ** row by row, to process output.      
   { | errmsg[ cell ] -- ior }
   errmsg[ off
   execute				\ xt was a SQL3Callback
   sql3-handle @ -rot NULL errmsg[ sqlite3_exec
   errmsg[ @ ?dup if sqlite3_free then
;
EXPORT db-process

: db-process-with			  \ z-addr xt1 xt2 -- ior
\ *G Execute SQL statement *\i{z-addr}.
\ ** When complete, word given by *\i{xt2} is repetidely called,
\ ** row by row, to process output.
\ ** *\i{ior} code signals operation result.    
\ ** *\i{xt1} is the execution token for a word that - when executed -
\ ** return a cell with client data.
\ ** Client data can be anything, but mostly will be an object handle
\ ** for OOP support (the *\fo{this} keyword).   
\ *C <clientData> VALUE clientDataPtr   \ xt1 is ' clientDataPtr.
\ *C : this ( -- handle) ... ;          \ xt1 is ' this   
   { z-addr xt1 xt2 | errmsg[ cell ] -- ior }
   errmsg[ off
   sql3-handle @ z-addr xt2 execute xt1 execute errmsg[ sqlite3_exec
   errmsg[ @ ?dup if sqlite3_free then
;
EXPORT db-process-with


: ior>excp				  \ ior -- excp|0
\ +G Map SQLite3 *\i{ior} codes from *\i{1..22} to VFX Forth exceptions.
\ +* All other values are mapped to 0.   
   dup 1 23 within 0= if drop 0 exit then
   1- ErrorBase +
;
   

: db-throw				  \ ior --
\ *G Map SQLite3 *\i{ior} into an appropiate VFX Forth exception and throw it.
   ior>excp throw
;
EXPORT db-throw


(( For DocGen Use. Defined at the bottom of the file.
: db-print				  \ z-addr --
\ *G Print the SQL statement *\i{z-addr} being executed and its result set 
\ ** nicely formatted.   
;
))

: SQLite3Callback:			  \ xt --
\ *G Declare a SQLite3 callback that, when called by the Linux C interface
\ ** it will execute in turn the action whose execution token is *\i{xt}
\ ** The action Forth word must have a signature like its C counterpart:
\ *C int action(void* cliData, int nCols, char* colValue[], char* colName[]);
   4 1 Callback:			\ create callback
   latest-xt execute set-callback	\ bind xt to recently created callback.
;
EXPORT SQLite3Callback:

: sql3-$@		       \ a-addr i -- ca u
\ *G Retrieve a string *\i{ca u} given the base array of pointers
\ ** *\i{a-addr} to Z-strings and the index into the array *\i{i}.
\ ** Intended to retrieve contents from the callback Forth word being used
\ ** to process rows in the ForthQL. See *\fo{SQLite3Callback:}.   
   cells + @ zcount
;
EXPORT sql3-$@

\ ==========================
\ *N SQLite3 Specific DB API
\ ==========================

\ *P This API export function for a non callback driven interface.
\ ** Iteration on the *\i{result set} is done by the application itself.

: sql3-prepare				  \ z-addr1 -- z-addr2 ior
\ *G Precompiles a SQL statement *\i{z-addr1} for later execution through one
\ ** or more calls to *\fo{SQL3-STEP}.
\ ** Only a single SQL statement is compiled, remaining
\ ** is left as *\i{z-addr2} for subsequent calls to *\fo{SQL3-PREPARE}.
\ ** An empty statement set *\fo{SQL3-STH} to zero.  
   { | z-addr2[ cell ] -- ior }			  
   sql3-handle @ swap zcount sql3-sth z-addr2[ sqlite3_prepare_v2
   z-addr2[ @ swap
; 
EXPORT sql3-prepare

: sql3-step        \ -- ior
\ *G Get next row from the result set.
\ ** *\i{ior} is *\fo{SQLITE_ROW} if result set not complete or
\ ** *\fo{SQLITE_DONE} otherwise. A different value signals an error.
   sql3-sth @ sqlite3_step
;
EXPORT sql3-step

: sql3-reset        \ -- ior
\ *G Reset the execution of a SQL statement.
   sql3-sth @ sqlite3_reset
;  
EXPORT sql3-reset

(( For backup and example purposes
: sql3-#rows				  \ -- n
   sql3-sth @ 0
   BEGIN over sqlite3_step SQLITE_ROW = WHILE 1+ REPEAT
   swap sqlite3_reset drop
;
))

: sql3-#cols       \ -- n
\ *G Return the number of columns in the result set. Only works after
\ ** calling *\fo{SQL3-PREPARE}.
   sql3-sth @ sqlite3_column_count
;
EXPORT sql3-#cols

: sql3-finalize    \ -- ior
\ *G Clean up the result set after being iterated.
\ ** Las thing to do after a *\fo{SQL3-PREPARE} and
\ ** zero or more *\fo{SQL3-STEP}s.   
\ ** *\i{ior} signals an error condition.   
   sql3-sth @ sqlite3_finalize  sql3-sth off

;
EXPORT sql3-finalize

: sql3-NULL	    \ -- ca1 u1
\ *G The verbatim *\f{NULL} string with its four characters.   
   s" NULL"
;
EXPORT sql3-NULL


: sql3-col-name    \ n -- ca u
\ *G Return the column name for column *\i{n}.
\ ** *\f{NULL} datatype is retrieved as the *\f{NULL} string.      
   sql3-sth @ SWAP sqlite3_column_name
   ?dup if zcount exit then
   sql3-NULL
;
EXPORT sql3-col-name

: sql3-col-text    \ n -- ca u
\ *G Return the column value as ASCII text for column *\i{n}.
\ ** *\f{NULL} datatype is retrieved as the *\f{NULL} string.   
   sql3-sth @ SWAP sqlite3_column_text
   ?dup if zcount exit then
   sql3-NULL
;
EXPORT sql3-col-text

: sql3-col-type   \ n1 -- n2
\ *G Return the column type for column *\i{n1} as defined in one of
\ ** the SQLITE constants *\fo{SQLITE_INTEGER} through *\fo{SQLITE_NULL}.
   sql3-sth @ SWAP sqlite3_column_type
;
EXPORT sql3-col-type 

: sql3-numeric?			  \ n1 -- flag
\ *G True if column type *\i{n1} is either *\fo{SQLITE_INTEGER}
\ ** or *\fo{SQLITE_FLOAT}.
   dup SQLITE_INTEGER = swap SQLITE_FLOAT = or
;
EXPORT sql3-numeric?


\ --------------------------
\ *H Pretty printing section
\ --------------------------

\ +P Dont need to be thread-safe,
\ +* only the operator's console thread needs this.

Variable SQL3-#ROWS
\ +G Performance counter. Keep the number of rows for
\ +* the printed result set for the final report.

Variable SQL3-T0
\ +G Performance counter. Start timestamp for the printed report.

: /perf
\ +G Initialize performance counters.
   SQL3-#ROWS off   ticks SQL3-T0 !
;


: #.##					  \ n1 --
\ +G Print time in seconds up two two decimal places given the input
\ +* time in 1/100th seconds.   
   0 <# # # [char] . hold #s #> type space
;

: .perf				  \ --
\ +G Print the summary of rows collected and how long did it take.
   SQL3-#ROWS ? ." rows in result set"
   ."  (" ticks SQL3-T0 @ -  10 / #.## ." sec)" 
   cr   
;

40 Constant MAXWIDTH
\ *G Maximun character width for a given column.

32 Constant MAXCOLS
\ *G Maximun number of columns to pretty-print.

MAXCOLS Array fwidth[	 
\ +G Static array of field widths for printing

8 Constant DEFAULT-FWIDTH
\ +G Default field width

: ?bounds	      \ n -- flag
\ +G Check for index out of bounds in column widths array.
\ +* Throw  *\fo{SQL3ColumnIndex} otherwise.       
   0 MAXCOLS within 0= SQL3ColumnIndex ?throw
;


: ?wide					  \ n1 --
\ +G Check for proper range of column width *\i{n1} between {1..MAXWIDTH}.
\ +* Throw *\fo{SQL3ColumnWidthError} otherwise.
   1 MAXWIDTH 1+ within 0= SQL3ColumnWidthError ?throw   
;


: th-colwidth				  \ n1 i --
\ *G Set the *\i{i}th column width to *\i{n1}.
\ ** Throw *\fo{SQL3ColumnIndex} if *\i{i} out of bounds or
\ ** *\fo{SQL3ColumnIndex} if *\i{n1} is too wide.
   dup ?bounds   over ?wide  fwidth[ !
;
EXPORT th-colwidth

: colwidth				  \ n1 --
\ *G Set all column widths to *\i{n1}.
\ ** Throw *\fo{SQL3ColumnIndex} if *\i{n1} is too wide. 
   dup ?wide
   MAXCOLS 0 DO
      dup i fwidth[ !
   LOOP
   drop
;
EXPORT colwidth

: +colwidth				  \ n1 --
\ *G Increment all column widths to *\i{n1}.
\ ** Throw *\fo{SQL3ColumnIndex} if the resulting width is too wide. 
   MAXCOLS 0 DO
      dup i fwidth[ @ + ?wide
      dup i fwidth[ +!
   LOOP
   drop
;
EXPORT +colwidth



: /fwidth     \ --
\ +G Initialize the field widths array to default size.
   MAXCOLS 0 ?do
      DEFAULT-FWIDTH i fwidth[ !
   loop
;

/fwidth

: dashes				  \ n --
\ +G Print *\i{n} times  *\fo{CHAR -} to standard output.   
   0 ?do [char] - emit loop
;


: plus
\ +G Print *\fo{CHAR +} to standard output.      
   [char] + emit
;

: pipe
\ +G Print *\fo{CHAR |} to standard output.   
   [char] | emit
;


: $type/				  \ ca1 u1 --
\ +G Print TEXT string with a final truncation *\fo{CHAR >} sign appended.
   type [char] > emit
;


: #type/				  \ ca1 u1 --
\ +G Print an unreadable string representing an INTEGER or FLOAT value
\ +* using a number of *\fo{CHAR #}.
   nip 0 ?do [char] # emit loop
;


: ltype					  \ ca1 u1 u2 --
\ +G Print a left justified string *\i{ca1 u1} in a field width of *\{u2}
\ +* filling with spaces the rest. Assumes *\f{u1 <= u2}.
   over - -rot type spaces
;


: rtype					  \ ca1 u1 u2 --
\ +G Print a right justified string *\i{ca1 u1} in a field width of *\{u2}
\ +* filling with spaces the rest. Assumes *\f{u1 <= u2}.
   over - spaces type
;


: .separator	       \ nCols --
\ +G Print the separator line between column names and values.
\ +* *\i{nCols} is the number of columns to print.   
   plus
   0 ?do
      1 dashes
      i fwidth[ @ dashes
      1 dashes plus
   loop
   cr
;


: .right&clipped			  \ ca1 u1 i -- Flag
\ +G Print and possibly clip a right justified string *\i{ca1 u1}
\ +* for column *\i{i}, taking into account the field width specifier
\ +* for that column.
\ +* Flag is true if clipped.      
   fwidth[ @
   2dup > if
      nip #type/ True			\ truncate string
   else
      rtype False			\ space + string
   then
;


: .left&clipped			          \ ca1 u1 i -- Flag
\ +G Print and possibly clip a left justified string *\i{ca1 u1}
\ +* for column *\i{i}, taking into account the field width specifier
\ +* for that column.
\ +* Flag is true if clipped.   
   fwidth[ @
   2dup > if
      nip $type/ True			\ truncate string
   else
      ltype False			\ string + space
   then
;


: .left-justified			  \ ca1 u1 i --
\ +G Like *\fo{.left&clipped} above but adding column decorations.   
   pipe space .left&clipped 0= if space then
;


: .right-justified
\ +G Like *\fo{.right&clipped} above but adding column decorations.      
   pipe space .right&clipped drop space
;


: .row	    \ nCols --
\ +G Print a single line of contents in result set.
\ +* Numeric values are right justified, all others left justified.
\ +* *\i{nCols} is the number of columns for each row.   
   0 ?do
      i dup sql3-col-text		\ -- i ca1 u1
      rot dup sql3-col-type		\ -- ca1 u1 i n1
      sql3-numeric? if
	 .right-justified
      else
	 .left-justified
      then
   loop
   pipe cr
;


: .contents				  \ nCols --
\ +G Print the query result set for the current SQL statement.
\ +* Also count the rows in the result set and store it into variable
\ +* *\fo{SQL3-#ROWS}.
\ +* *\i{nCols} is the number of columns for each row.      
   sql3-sth @  0 
   BEGIN
      over sqlite3_step SQLITE_ROW =
   WHILE				\ -- nCols sth nRows
	 rot dup .row -rot 1+
   REPEAT
   SQL3-#ROWS !
   2drop
;


: .colnames				  \ nCols --
\ +G Print the query column names.
\ +* *\i{nCols} is the number of columns for each row.
   0 ?do
      i dup sql3-col-name rot .left-justified
   loop
   pipe cr
;


: .header		\ nCols --
\ +G Print the query header: separators and column names, perhaps clipped.
\ +* *\i{nCols} is the number of columns for each row.         
   dup .separator dup .colnames .separator 
;

: (.sql3-rows)				  \ --
\ +G Do the actual pretty-print of the result set for a single SQL statement
\ +* when all conditions have been met.
   sql3-#cols ?dup if  
      dup ?bounds dup .header dup .contents .separator
   else
      sql3-step db-throw  \ sql3-finalize wants this even for empty stmt.
   then
;

\ ================================================
\ THIS IS COPIED FROM MY XML UTILITIES
\ I MAY PUT IT IN THE EXTRAS PROJECT AFTER ALL ...
\ ================================================

: $1-				  \  ca1 u1 -- ca1-1 u1-1
\ +G Trim right string by one position.
\ +* *\i{ca1} is the *\b{string end address}, *\i{u1} its length
   1- swap 1 chars - swap ;			


: white?	      \ c -- flag   
\ +G True if *\i{c} is a whitespace-class character.
\ +* This means ASCII codes less or equal than 32.
\ +* So we conveniently skip spaces, TABs, LFs and CRs.
   BL > 0=
;


: -leader			 \ ca1 u1 -- ca2 u2
\ +G trims string to skip leading spaces, TABs, LFs and CRs,
\ +* adjusting length as necessary.
\ +* VFX's *\fo{-LEADING} is similar but does not skip TABs, LFs or CRs. 
   begin  over c@ white? over 0<> and while 1 /string repeat
;


: -trailer		       \ ca1 u1 -- ca2 u2 
\ +G trims string to skip trailing spaces, TABs, LFs and CRs,
\ +* adjusting length as necessary.
\ +* VFX's *\fo{-TRAILING} is similar but does not skip TABs, LFs or CRs. 
   2dup chars + 1 chars - swap		\ produce ending addr and length
   begin over c@ white? over 0<> and while $1- repeat nip
;

\ ================================================

: .sql3-statement				  \ z-addr1 z-addr2 --
\ +G Extract and print the SQL statement about to execute in one or more
\ +* *\fo{SQL3-STEP}s given the old and new positions of the SQL buffer.
\ +* *\i{z-addr1} is the old position and  *\i{z-addr2} is the new one
\ +* being returned by *\fo{SQL3-PREPARE}.   
\ +* Assume *\i{z-addr2} > *\i{z-addr1}.
   over - -leader -trailer type cr
;


: .sql3-rows			        \ z-addr1 -- z-addr2 
\ +G Pretty-print the result set for a single SQL statement.
\ +* For an statement not yielding a result set, only prints the statistics.   
\ +* Leave the remaining SQL to be executed at *\i{z-addr2}.
\ +* If no more statements remain, *\i{z-addr2} is NULL.   
   /perf dup
   sql3-prepare db-throw 		\ -- z-addr1 z-addr2
   sql3-sth @ 0= if 2drop NULL exit then \ No more statements found.
   tuck .sql3-statement (.sql3-rows)
   sql3-finalize db-throw
   .perf cr
;



: db-print				  \ z-addr --
\ Print the SQL statement *\i{z-addr} being executed and its result set
\ nicely formatted.
   cr BEGIN dup WHILE .sql3-rows REPEAT drop
;
EXPORT db-print

\ +H SQL-RESULT Callback

\ +P Example using callbacks to process data row by row.
\ +* We have the limitation that SQLite3 does not notify when the
\ +* last row in the result set is reached.
\ +* Also, SQLite3 callback only gets data as strings and not native
\ +* data types.

Variable preamble			\ Flag: Things to do fist time
preamble on

: .myrow	    \ colValues* nCols --
\ +G Print a single line of contents in result set.
\ +* All values are left justified
\ +* *\i{nCols} is the number of columns for each row.
\ +* *\i{colValues} is an array of pointers to strings.   
   0 ?do
      i 2dup				\ array i array i
      sql3-$@				\ array i ca1 u1
      rot .left-justified
   loop
   drop
   pipe cr
;


: .myheader				  \ colNames* nCols --
\ +G Print the header only once.
\ +* All values are left justified
\ +* *\i{nCols} is the number of columns for each row.
\ +* *\i{colNames} is an array of pointers to strings.      
   preamble @off if
      cr tuck .separator  over .myrow  .separator
   else
      2drop
   then
;


: (SQL-RESULT)   { client* nCols colValues* colNames* -- ior }
\ +G Forth word that is called on every row in the query result set.   
   colNames*  nCols .myheader
   colValues* nCols .myrow
   SQLITE_OK
;


' (SQL-RESULT)
SQLite3Callback: SQL-RESULT      \ client* nCols colValues* colNames* -- ior 
\ *G Print a query result set, like *\fo{|SQL.} but done as
\ ** *\fo{|SQL> SQL-RESULT}. Printed formatting has deficiencies.
\ ** Use *\fo{|SQL.} instead.

EXPORT SQL-RESULT

END-MODULE
